import json
from concurrent import futures

import grpc
import pydash

import llnwLDS_pb2_grpc
import llnw_pb2_grpc
from ll_sdk.config_api import ConfigApiClient
from llnw_pb2 import *
from llnwLDS_pb2 import *


def get_cl(base):
    return ConfigApiClient('apis.llnw.com', base.username, base.key)


def delivery_to_grpc(delivery):
    return DeliveryEntity(
        uuid=delivery.get('uuid', None),
        isEnabled=delivery.get('isEnabled', None),
        shortname=delivery.get('shortname', None),
        status=pydash.objects.get(delivery, "status.state"),
        sourceHostname=pydash.objects.get(delivery, "body.sourceHostname"),
        publishedHostname=pydash.objects.get(delivery, "body.publishedHostname"),
        sourceUrlPath=pydash.objects.get(delivery, "body.sourceUrlPath"),
        publishedUrlPath=pydash.objects.get(delivery, "body.publishedUrlPath"),
        protocolSets=[ProtocolSet(
            sourceProtocol=protocolSets.get('sourceProtocol', None),
            publishedProtocol=protocolSets.get('publishedProtocol', None),
            options=[ProtocolSetOptions(
                name=options.get('name', None),
                parameters=[str(param) for param in options.get('parameters', None)],
                source=options.get('source', None),
            ) for options in protocolSets.get('options', [])]
        ) for protocolSets in pydash.objects.get(delivery, 'body.protocolSets', [])]
    )


def delivery_from_grpc(delivery_entity: DeliveryEntity):
    return {
        "uuid": delivery_entity.uuid,
        "accounts": [
            {
                "shortname": delivery_entity.shortname
            }
        ],
        "body": {
            "description": "由 python API 提交",
            "protocolSets": [
                {
                    "options": [
                        {
                            "name": po.name,
                            "parameters": [
                                para if po.name != "refresh_absmin" and po.name != "refresh_absmax" else int(para)
                                for para in po.parameters],
                            "source": po.source
                        } for po in protocol.options
                    ],
                    "publishedProtocol": protocol.publishedProtocol,
                    "sourceProtocol": protocol.sourceProtocol
                } for protocol in delivery_entity.protocolSets
            ],
            "publishedHostname": delivery_entity.publishedHostname,
            "publishedUrlPath": delivery_entity.publishedUrlPath,
            "serviceKey": {
                "name": "delivery"
            },
            "serviceProfileName": "LLNW-Generic",
            "serviceProfileVersion": 2,
            "sourceHostname": delivery_entity.sourceHostname,
            "sourceUrlPath": delivery_entity.sourceUrlPath
        },
        "meta": {
            "allowManualDuplicate": True
        }
    }


def lds_config_from_grpc(request):
    storage = {
        "accessKey": request.accessKey,
        "bucketName": request.bucketName,
        "location": request.location,
        "path": request.path,
        "region": request.region,
        "secretKey": request.secretKey
    }
    return {
        "compressionType": "lz4",
        "isActive": "true",
        "logFields": ["sc-request-id", "s-host", "datetime", "c-country-code", "sc-status", "c-ip", "cs-uri",
                      "sc-bytes"],
        "name": request.name,
        "serviceType": "http",
        "storage": storage
    }


def lds_uuids_to_grpc(res):
    uuids = []
    results = res.get("results")
    if results is None:
        return ListLDSResponse(uuids=uuids)
    else:
        for result in results:
            uuids.append(result.get("uuid"))
        return ListLDSResponse(uuids=uuids)


def check_error(res, context):
    if not res.get('success', True):
        match res.get('errorType'):
            case 'validation':
                context.abort(grpc.StatusCode.INVALID_ARGUMENT, json.dumps(res.get('errors')))
            case _:
                context.abort(grpc.StatusCode.UNKNOWN, json.dumps(res))


def check_lds_create_result(res):
    if res.get('isEnabled') is None:
        return CreateLDSResponse(isSuccess=False)
    else:
        if res.get('isEnabled') is True:
            return CreateLDSResponse(isSuccess=True)
        else:
            return CreateLDSResponse(isSuccess=False)


def check_lds_delete_result(res_code):
    print("响应码：", res_code.status_code)
    if res_code.status_code == 200:
        return DeleteLDSResponse(isSuccess=True)
    else:
        return DeleteLDSResponse(isSuccess=False)


class DeliveryServiceServicer(llnw_pb2_grpc.DeliveryServiceServicer):

    def ListDelivery(self, request, context):
        res = get_cl(request.base).list_delivery_service_instances(request.shortName).json()
        check_error(res, context)
        return DeliveryList(deliveries=[delivery_to_grpc(results) for results in res['results']])

    def ValidateDelivery(self, request, context):
        delivery_config = delivery_from_grpc(request.delivery)
        del delivery_config['uuid']
        res = get_cl(request.base).validate_delivery_service_instance(delivery_config).json()
        check_error(res, context)
        return delivery_to_grpc(res)

    def CreateDelivery(self, request, context):
        delivery_config = delivery_from_grpc(request.delivery)
        del delivery_config['uuid']
        res = get_cl(request.base).create_delivery_service_instance(delivery_config).json()
        check_error(res, context)
        return delivery_to_grpc(res)

    def GetDelivery(self, request, context):
        res = get_cl(request.base).get_delivery_service_instance(request.id).json()
        check_error(res, context)
        return delivery_to_grpc(res)

    def UpdateDelivery(self, request, context):
        delivery_config = delivery_from_grpc(request.delivery)
        res = get_cl(request.base).update_delivery_service_instance(delivery_config['uuid'], delivery_config).json()
        check_error(res, context)
        return delivery_to_grpc(res)

    def DeleteDelivery(self, request, context):
        res = get_cl(request.base).delete_delivery_service_instance(request.id).json()
        check_error(res, context)
        return delivery_to_grpc(res)


class LogDeliveryServiceServicer(llnwLDS_pb2_grpc.LogDeliveryServiceServicer):

    def CreateLDS(self, request, context):
        print("CreateLDS")
        print("请求：", request)
        try:
            lds_config = lds_config_from_grpc(request)
            res = get_cl(request.base).create_lds(shortname=request.shortName, config={"body": lds_config}).json()
            print("响应：", res)
        except:
            return CreateLDSResponse(isSuccess=False)
        else:
            return check_lds_create_result(res)

    def DeleteLDS(self, request, context):
        print("DeleteLDS")
        print("请求：", request)
        try:
            res_code = get_cl(request.base).delete_lds(shortname=request.shortName, uuid=request.uuid)
        except:
            return DeleteLDSResponse(isSuccess=False)
        else:
            return check_lds_delete_result(res_code)

    def ListLDS(self, request, context):
        print("ListLDS")
        print("请求：", request)
        try:
            res = get_cl(request.base).list_lds(shortname=request.shortName).json()
            print("响应：", res)
        except:
            return ListLDSResponse(uuids=[])
        else:
            return lds_uuids_to_grpc(res)


if __name__ == '__main__':
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
    llnw_pb2_grpc.add_DeliveryServiceServicer_to_server(DeliveryServiceServicer(), server)
    llnwLDS_pb2_grpc.add_LogDeliveryServiceServicer_to_server(LogDeliveryServiceServicer(), server)
    server.add_insecure_port('[::]:50052')
    server.start()
    server.wait_for_termination()
